/////////////////////////////////////////////////////////////////////////
//	"devtest_dev.cpp"
//------------------------------------------------------------------------
//	$Id: devtest_dev.cpp 145 2005-12-19 06:03:27Z taro $
//////////////////////////////////////////////////////////////////////////

#include	<stdlib.h>
#include	<stdio.h>
#include	<string.h>
#include	<ctype.h>
#if	defined(LINUX)
# include	"linux/hdreg.h"
#endif
#include	"devtest.h"
#include	"defs.h"

//////////////////////////////////////////////////////////////////////////
//	Print Device Information

int		DevTest::PrintDevInfo(DevTest::DevParam & devParam)
{
	unsigned					pathId, targetId, lun;
	char						t1[256], t2[256];

	if ( devParam.CheckInqData() == FALSE )
		return -1;
	pathId   = devParam.inqData->GetPathId();
	targetId = devParam.inqData->GetTargetId();
	lun      = devParam.inqData->GetLun();

#if	defined(WIN32)
	{
		IO_SCSI_CAPABILITIES	capa;

		if ( this->devCtl.ScsiGetCapabilities(capa) == 0 && capa.Length >= 24 )
		{
			//  ----+----1----+----2----+----3----+----4----+----5----+----6----+----7----+----8
			// "    Disk Capabilities                ##########"
			// "    <-----------------------------------------> ##########"
			PutFmt(4, "Disk Capabilities"            , "", "");
			PutFmt(8, "Maximum Transfer Length"      , Fmt(t2,"%u",(unsigned)capa.MaximumTransferLength      ), "");
			PutFmt(8, "Maximum Physical Pages"       , Fmt(t2,"%u",(unsigned)capa.MaximumPhysicalPages       ), "");
			PutFmt(8, "Supported Asynchronous Events", Fmt(t2,"%u",(unsigned)capa.SupportedAsynchronousEvents), "");
			PutFmt(8, "Alignment Mask"               , Fmt(t2,"%u",(unsigned)capa.AlignmentMask              ), "");
			PutFmt(8, "Tagged Queuing"               , Fmt(t2,"%u",(unsigned)capa.TaggedQueuing    ? 1 : 0   ), "");
			PutFmt(8, "Adapter Scans Down"           , Fmt(t2,"%u",(unsigned)capa.AdapterScansDown ? 1 : 0   ), "");
			PutFmt(8, "Adapter Uses PIO"             , Fmt(t2,"%u",(unsigned)capa.AdapterUsesPio   ? 1 : 0   ), "");
		}
	}
#endif
#if	defined(WIN32)
	{
		GET_LENGTH_INFORMATION	lenInfo;

		if ( this->devCtl.DiskGetLengthInfo(lenInfo) == 0 )
		{
			PutPos(4         , "Disk Length Information",
			       COLUMN3-21, Fmt(t1,"%20,I64u", (UINT64)lenInfo.Length.QuadPart),
			       COLUMN3   , Fmt(t2, (this->opt.byteUnit == BYTEUNIT_210 ? "(%I64K)":"(%I64k)"), (UINT64)lenInfo.Length.QuadPart),
			       0);
		}
	}
	{
		DISK_GEOMETRY	geo;

		if ( this->devCtl.DiskGetDriveGeometry(geo) == 0 )
		{
			PutPos(4        , "Disk Drive Geometry",
			       COLUMN2-4, Fmt(t1,"Media Type:%02XH", (unsigned)geo.MediaType),
			       COLUMN3  , this->devCtl.GetMediaTypeName(geo.MediaType),
			       0);
			sprintf(t1," %"FI64"u/%u/%u : %u"
			  ,    (INT64)geo.Cylinders.QuadPart 
			  , (unsigned)geo.TracksPerCylinder
			  , (unsigned)geo.SectorsPerTrack  
			  , (unsigned)geo.BytesPerSector   
			);
			PutPos(8, "Cylinder/Track/Sector:Size"
			  , COLUMN2 + 10 - strlen(t1), t1
			  , COLUMN3                  , Fmt(t2, " (%,I64u)", geo.Cylinders.QuadPart * geo.TracksPerCylinder * geo.SectorsPerTrack * geo.BytesPerSector)
			  , 0);
		}
	}
	{
		PARTITION_INFORMATION	partInfo;

		if ( this->devCtl.DiskGetPartitionInfo(partInfo) == 0 )
		{
			PutPos(4         , "Disk Partition Information",
			       COLUMN3-18, Fmt(t1,"Partition Type:%02X", partInfo.PartitionType),
			       COLUMN3   , DevTest::GetPartitionType(partInfo.PartitionType),
			       0);
			PutPos(8         , "Starting Offset",
			       COLUMN3-21, Fmt(t1,"%20,I64u", (UINT64)partInfo.StartingOffset.QuadPart),
			       COLUMN3   , Fmt(t2, (this->opt.byteUnit == BYTEUNIT_210 ? "(%I64K)":"(%I64k)"), (UINT64)partInfo.StartingOffset.QuadPart),
			       0);
			PutPos(8         , "Partition Length",
			       COLUMN3-21, Fmt(t1,"%20,I64u", (UINT64)partInfo.PartitionLength.QuadPart),
			       COLUMN3   , Fmt(t2, (this->opt.byteUnit == BYTEUNIT_210 ? "(%I64K)":"(%I64k)"), (UINT64)partInfo.PartitionLength.QuadPart),
			       0);
			PutPos(8         , "Hidden Sectors",
			       COLUMN3-21, Fmt(t1,"%20,u", (unsigned)partInfo.HiddenSectors),
			       0);
			PutPos(8         , "Partition Number",
			       COLUMN3-21, Fmt(t1,"%20,u", (unsigned)partInfo.PartitionNumber),
			       0);
		}
	}
#endif
#if	defined(LINUX)
	if ( this->devCtl.GetHType() == DevCtl::HT_BLKDEV
	  || this->devCtl.GetHType() == DevCtl::HT_CDROM
	 )
	{
		struct hd_driveid		drvid;
		char					*t;

		memset( &drvid, 0, sizeof(drvid));
		if ( ioctl(this->devCtl.GetHandle(), HDIO_GET_IDENTITY, &drvid) == 0 )
		{
			PutPos(4        , "HDIO GET IDENTITY", 0);
			if ( drvid.cyls || drvid.heads || drvid.sectors )
			{
				sprintf(t1, "%u/%u/%u", (unsigned)drvid.cyls, (unsigned)drvid.heads, (unsigned)drvid.sectors);
				PutPos(8, "Physical Cylinder/Heads/Sectors", COLUMN2-6, Fmt(t2,"%16s",t1), 0);
			}
			if ( drvid.track_bytes  ) PutPos(8, "Unformatted Track size" , COLUMN2, Fmt(t1, "%10u", (unsigned)drvid.track_bytes ) , 0);
			if ( drvid.sector_bytes ) PutPos(8, "Unformatted Sector size", COLUMN2, Fmt(t1, "%10u", (unsigned)drvid.sector_bytes) , 0);
			if ( drvid.serial_no[0] ) PutPos(8, "Serial No."             , COLUMN2-10, Fmt(t1, "%20.20s", drvid.serial_no       ) , 0);
			if ( drvid.ecc_bytes    ) PutPos(8, "ECC Bytes"              , COLUMN2, Fmt(t1, "%10u", (unsigned)drvid.ecc_bytes   ) , 0);
			if ( drvid.fw_rev[0]    ) PutPos(8, "F/W Revision"           , COLUMN2+ 2, Fmt(t1, "%8.8s", drvid.fw_rev            ) , 0);
			if ( drvid.model[0]     ) PutPos(8, "Model"                  , COLUMN2-30, Fmt(t1, "%40.40s", drvid.model           ) , 0);
			switch ( drvid.tPIO )
			{
				case 0 : t = "slow"  ; break;
				case 1 : t = "medium"; break;
				case 2 : t = "fast"  ; break;
				default: sprintf(t1, "%02X", (unsigned)drvid.tPIO); t = t1; break;
			}
			PutPos(8, "PIO"                  , COLUMN2, Fmt(t1, "%10.10s", t) , 0);
			switch ( drvid.tDMA )
			{
				case 0 : t = "slow"  ; break;
				case 1 : t = "medium"; break;
				case 2 : t = "fast"  ; break;
				default: sprintf(t1, "%02X", (unsigned)drvid.tDMA); t = t1; break;
			}
			PutPos(8, "DMA"                  , COLUMN2, Fmt(t1, "%10.10s", t) , 0);
			if ( drvid.cur_cyls || drvid.cur_heads || drvid.cur_sectors )
			{
				sprintf(t1, "%u/%u/%u", (unsigned)drvid.cur_cyls, (unsigned)drvid.cur_heads, (unsigned)drvid.cur_sectors);
				PutPos(8, "Logical Cylinder/Heads/Sectors", COLUMN2-6, Fmt(t2,"%16s",t1), 0);
			}
			if ( drvid.lba_capacity ) PutPos(8, "LBA Capacity"           , COLUMN2, Fmt(t1, "%10u", (unsigned)drvid.lba_capacity) , 0);
		}
	}
#endif
	{
		BOOL	check = FALSE;

		if ( this->devCtl.ScsiTestUnitReady(pathId, targetId, lun) == 0 )
		{
			PutFmt(4, "Test Unit Ready", "OK", "");
			check = TRUE;
		}
		if ( devParam.inqData->unknownId == FALSE )
			check = TRUE;
		if ( check )
		{
			this->PrintDevInfo_vpd(devParam);
			this->PrintDevInfo_modeSense(devParam);
			this->PrintDevInfo_defectData(devParam);
			this->PrintDevInfo_getConfiguration(devParam);
			this->PrintDevInfo_dvdStructure(devParam);
		}
	}
	return 0;
}

#define	DEVTEST_VD_VPD	// Use Dummy VPD Data

int		DevTest::PrintDevInfo_vpd(DevTest::DevParam & devParam)
{
	unsigned	pathId, targetId, lun;
	BYTE		buffer[255], codelist[255], map[256];
	unsigned	cl_num, cl_pos;
	enum map_enu { MAP_OFF, MAP_DISP, MAP_ERR };
	DevCtl::SCSI_VPD_PAGECODE_LIST	*pcl;

	if ( devParam.CheckInqData() == FALSE )
		return -1;
	pathId   = devParam.inqData->GetPathId();
	targetId = devParam.inqData->GetTargetId();
	lun      = devParam.inqData->GetLun();
	memset(map, MAP_OFF, sizeof(map));

	// Get PageCode List
	if ( this->devCtl.ScsiInquiry(pathId, targetId, lun, 0x00, sizeof(buffer), buffer) < 0 )
		return -1;	// error
	pcl = (DevCtl::SCSI_VPD_PAGECODE_LIST *) buffer;
	if ( pcl->pagecode != 0x00 || pcl->pagelen == 0 )
		return -1;	// error
	cl_num = pcl->pagelen;
	memcpy(codelist, buffer + sizeof(*pcl), cl_num);
	map[0x00] = MAP_DISP;
	for ( cl_pos = 0; cl_pos < cl_num; cl_pos++ )
	{
		unsigned	page, len;
		BYTE		*p;

		page = codelist[cl_pos];
		if ( map[page] == MAP_DISP )
			continue;
		memset( buffer, 0, sizeof(buffer));
		if ( this->devCtl.ScsiInquiry(pathId, targetId, lun, page, sizeof(buffer), buffer) < 0 )
			break;	// error
		if ( page != buffer[1] )
			page = buffer[1];	// error ?
		len = buffer[3];
		p = buffer + 4;
		if ( len > 0 && map[page] != MAP_DISP )
		{
			char	msg[256];

			//  ----+----1----+----2----+----3----+----4----+----5----+----6----+----7----+----8
			// "    VPD:Page Code XXH (length:###)
			// "                                        ####: ## ## ## ## ## ## ## ##  ........
			this->PrintDumpOct(Fmt(msg,"VPD:Page Code %02XH (length:%3u)", page, len), TRUE, len, p);
			map[page] = MAP_DISP;
		}
	}
	return 0;
}

int		DevTest::PrintDevInfo_modeSense(DevTest::DevParam & devParam)
{
	unsigned	pathId, targetId, lun, mlen = 0, bdlen = 0;
	BYTE		buffer[4096], *p, *end;
	BOOL		cdb10 = this->opt.mode_cdb == MODE_CDB10 ? TRUE : FALSE;
	DevCtl::SCSI_MODE_PARAMETER_HEADER6		*mhead6;
	DevCtl::SCSI_MODE_PARAMETER_HEADER10	*mhead10;

	if ( devParam.CheckInqData() == FALSE )
		return -1;
	pathId   = devParam.inqData->GetPathId();
	targetId = devParam.inqData->GetTargetId();
	lun      = devParam.inqData->GetLun();
	for (;;)
	{
		memset(buffer, 0, sizeof(buffer));
		if ( cdb10 )
		{
			if ( this->devCtl.ScsiModeSense(pathId, targetId, lun, FALSE, 0, 0x3F, sizeof(buffer), buffer) >= 0 )
			{
				mhead10 = (DevCtl::SCSI_MODE_PARAMETER_HEADER10 *) buffer;
				mlen  = SC16(mhead10->length    );
				bdlen = SC16(mhead10->blkDescLen);
			} else
				return -1;	// error
		} else
		{
			if ( this->devCtl.ScsiModeSense(pathId, targetId, lun, FALSE, 0, 0x3F, 255, buffer) >= 0 )
			{
				mhead6 = (DevCtl::SCSI_MODE_PARAMETER_HEADER6 *) buffer;
				mlen  = mhead6->length;
				bdlen = mhead6->blkDescLen;
				p = buffer + sizeof(*mhead6) + bdlen;
				if ( mlen > 6 && p[0] == 0 && p[1] == 0 )
				{	// modeSense ucess, but data nothing
					cdb10 = TRUE;
					continue;
				}
			} else
			{
				cdb10 = TRUE;
				continue;	// retry
			}
		}
		break;
	}

	if ( mlen == 0 || mlen > sizeof(buffer) )
		return 0;
	if ( this->opt.dumpModeSense )
	{
		if ( cdb10 )
			this->PrintDumpOct("modeSense", TRUE, mlen+sizeof(mhead10->length), buffer);
		else
			this->PrintDumpOct("modeSense", TRUE, mlen+sizeof(mhead6->length), buffer);
	}
#if	defined(DEVTEST_DUMP_CSTYLE)
	if ( this->opt.dumpCStyle )
	{
		if ( cdb10 )
		{
			fprintf(stdout, "modeSense (10)\n");
			this->PrintDumpCStyle( mlen+sizeof(mhead10->length), buffer );
		} else
		{
			fprintf(stdout, "modeSense (6)\n");
			this->PrintDumpCStyle( mlen+sizeof(mhead6->length), buffer );
		}
	}
#endif
	if ( bdlen > 0 )
	{	// block descripter
		char				text[80];

		if ( cdb10 )
		{
			p = buffer + sizeof(*mhead10);
		} else
		{
			p = buffer + sizeof(*mhead6);
		}
		//  ----+----1----+----2----+----3----+----4----+----5----+----6----+----7----+----8
		// "    modeSense: Block Descripter ##H  (########) ##########"
		// "               %-33s %10u\n", "Write Protect", (buffer[2]>>7) & 1);
		sprintf(text, "Block Descripter %02XH", p[0]);
		fprintf(stdout, "    modeSense: %-22s (%8u) %10u\n"
		  , text
		  , SC24(p+5)
		  , SC24(p+1)
		);
	}
	{
		BYTE				*data;
		if ( (data = devParam.inqData->GetData()) != NULL )
		{
			switch ( data[1] & 0x3F )
			{
				case 0x00:	// direct access device
				{
					unsigned	wp, dpofua;

					if ( cdb10 )
					{
						wp     = (mhead10->devSpcParam>>7) & 1;
						dpofua = (mhead10->devSpcParam>>4) & 1;
					} else
					{
						wp     = (mhead6->devSpcParam>>7) & 1;
						dpofua = (mhead6->devSpcParam>>4) & 1;
					}
					fprintf(stdout, "    modeSense: %-33s %10u\n", "Write Protect", wp);
					fprintf(stdout, "               %-33s %10u\n", "Disable PageOut/Force Unit Access", dpofua);
					break;
				}
			}
		}
	}
	if ( cdb10 )
	{
		p = buffer + sizeof(*mhead10) + bdlen;
		end = buffer + sizeof(mhead10->length) + mlen;
	} else
	{
		p = buffer + sizeof(*mhead6) + bdlen;
		end = buffer + sizeof(mhead6->length) + mlen;
	}

	while ( p < end )
	{
		char	msg[80], s1[80], t1[256];

		if ( p[1] == 0 )
			break;
		//  ----+----1----+----2----+----3----+----4----+----5----+----6----+----7----+----8
		// "    modeSense: pagecode ##H size:%3u - Disconnect/Reconnect"
		sprintf(msg, "modeSense: pagecode %02XH size:%3u", p[0] & 0x3F, p[1]);
		switch ( p[0] & 0x3F )
		{
			case 0x01:	// Read/Write Error Recovery Parameter
			{
				//  ----+----1----+----2----+----3----+----4----+----5----+----6----+----7----+----8
				// "    modeSense: pagecode ##H - Disconnect/Reconnect"
				// "    <-----------------------------------------> ##########"
				if ( p[1] < 10 )
					goto ERR;
				fprintf(stdout, "    %-33s - Read/Write Error Recovery Parameter\n", msg);
				PutPos(15, "Automatic Write Reallocation Enabled" , COLUMN2+5, Fmt(t1,"%5s", GetOnOff(p[2] & 0x80)), 0);
				PutPos(15, "Automatic Read  Reallocation Enabled" , COLUMN2+5, Fmt(t1,"%5s", GetOnOff(p[2] & 0x40)), 0);
				PutPos(15, "Transfer Block"                       , COLUMN2+5, Fmt(t1,"%5s", GetOnOff(p[2] & 0x20)), 0);
				PutPos(15, "Read Continuous"                      , COLUMN2+5, Fmt(t1,"%5s", GetOnOff(p[2] & 0x10)), 0);
				PutPos(15, "Enable Early Recovery"                , COLUMN2+5, Fmt(t1,"%5s", GetOnOff(p[2] & 0x08)), 0);
				PutPos(15, "Post Error"                           , COLUMN2+5, Fmt(t1,"%5s", GetOnOff(p[2] & 0x04)), 0);
				PutPos(15, "Disable Tranfer on Error"             , COLUMN2+5, Fmt(t1,"%5s", GetOnOff(p[2] & 0x02)), 0);
				PutPos(15, "Disable Correction"                   , COLUMN2+5, Fmt(t1,"%5s", GetOnOff(p[2] & 0x01)), 0);
				PutPos(15, "Read Retry Count"                     , COLUMN2  , Fmt(t1,"%10u", p[3]), 0);
				PutPos(15, "Correction Span"                      , COLUMN2  , Fmt(t1,"%10u", p[4]), 0);
				PutPos(15, "Head Offset"                          , COLUMN2  , Fmt(t1,"%10d", (int)((signed char)p[5])), 0);
				PutPos(15, "Data Strobe Offset"                   , COLUMN2  , Fmt(t1,"%10d", (int)((signed char)p[6])), 0);
				PutPos(15, "Write Retry Count"                    , COLUMN2  , Fmt(t1,"%10u", p[8]), 0);
				PutPos(15, "Recovery Time Limit"                  , COLUMN2  , Fmt(t1,"%10u ms", SC16(p+10)), 0);
				break;
			}
			case 0x02:
			{
				DevCtl::ScsiPagecode02_Drm	*drm;

				if ( p[1] + 2 < 16 )
					goto ERR;
				fprintf(stdout, "    %-33s - Disconnect/Reconnect\n", msg);
				drm = (DevCtl::ScsiPagecode02_Drm *) p;
				PutPos(15,"Buffer Full Ratio"               , COLUMN2, Fmt(t1,"%10u",drm->buf_full_ratio         ), 0);
				PutPos(15,"Buffer Empty Ratio"              , COLUMN2, Fmt(t1,"%10u",drm->buf_empty_ratio        ), 0);
				PutPos(15,"Bus Inactive Limit"              , COLUMN2, Fmt(t1,"%10u",SC16(drm->bus_inact_limit  )), 0);
				PutPos(15,"Disconnect Time Limit"           , COLUMN2, Fmt(t1,"%10u",SC16(drm->disconn_timelimit)), 0);
				PutPos(15,"Connect Time Limit"              , COLUMN2, Fmt(t1,"%10u",SC16(drm->connect_timelimit)), 0);
				PutPos(15,"Maximum Burst Size"              , COLUMN2, Fmt(t1,"%10u",SC16(drm->max_burst_size   )), 0);
				PutPos(15,"Data Transfer Disconnect Control", COLUMN2, Fmt(t1,"%10u",drm->dtdc                   ), 0);
				PutPos(15,"Disconnect Immediate"            , COLUMN2, Fmt(t1,"%10s",GetOnOff(drm->dimm         )), 0);
				PutPos(15,"Fair Arbitration"                , COLUMN2, Fmt(t1,"%10u",drm->fair_arbitration       ), 0);
				PutPos(15,"Enable Modify Data Pointers"     , COLUMN2, Fmt(t1,"%10s",GetOnOff(drm->emdp         )), 0);
				PutPos(15,"First Burst Size"                , COLUMN2, Fmt(t1,"%10u",SC16(drm->first_burst_size )), 0);
				break;
			}
			case 0x04:	// Drive Parameter
			{
				static const char * rpl[] =
				{
					"Sync: disable or not supported",
					"Sync: slave",
					"Sync: master",
					"Sync: master control",
				};
				if ( p[1] < 22 )
					goto ERR;
				PutPos(4,msg, 38," - Drive Parameter", 0);
				PutPos(15, "Number of Cylinders"                    , COLUMN2  , Fmt(t1,"%10u", (p[2]<<16)+(p[3]<<8)+p[4]), 0);
				PutPos(15, "Number of Heads"                        , COLUMN2  , Fmt(t1,"%10u", p[5]), 0);
				PutPos(15, "Starting Cylinder-Write Precompensation", COLUMN2  , Fmt(t1,"%10u", (p[6]<<16)+(p[7]<<8)+p[8]), 0);
				PutPos(15, "Starting Cylinder-Reduced Write Current", COLUMN2  , Fmt(t1,"%10u", (p[9]<<16)+(p[10]<<8)+p[11]), 0);
				PutPos(15, "Drive Step Rate"                        , COLUMN2  , Fmt(t1,"%10u", (p[12]<<8)+p[13]), 0);
				PutPos(15, "Landing Zone Cylinder"                  , COLUMN2  , Fmt(t1,"%10u", (p[14]<<16)+(p[15]<<8)+p[16]), 0);
				PutPos(15, "Rotational Position Locking"            , COLUMN2  , Fmt(t1,"%10u", p[17] & 3), COLUMN2 + 11, rpl[p[17]&3], 0);
				PutPos(15, "Rotational offset"                      , COLUMN2  , Fmt(t1,"%10d", (int)((signed char)p[18])), 0);
				PutPos(15, "Medium rotation rate"                   , COLUMN2  , Fmt(t1,"%10u", (p[20]<<8)+p[21]), 0);
				break;
			}
			case 0x05:	// Flexible Disk Mode page
			{
				DevCtl::ScsiPagecode05_Fdm	*fdm;
				unsigned					len;
				const char					*s;

				if ( p[1] + 2 < 10 )	// 32
					goto ERR;
				fdm = (DevCtl::ScsiPagecode05_Fdm *) p;
				fprintf(stdout, "    %-33s - Flexible Disk Mode\n", msg);
				PutPos(15, "Transfer Rate"                    , COLUMN2  , Fmt(t1,"%10ukbit/s", (fdm->transfer_rate[0]<<8)+fdm->transfer_rate[1]), 0);
				sprintf(t1, "%u/%u/%u %u"
				  , (fdm->cylinders[0]<<8) + fdm->cylinders[1]
				  , fdm->sectors
				  , fdm->heads
				  , (fdm->sector_size[0]<<8) + fdm->sector_size[1]
				);
				len = strlen(t1);
				PutPos(15, "Cyl/Sec/Head sector size", COLUMN2+10-len, t1, 0);
			if ( fdm->length + 2 >= sizeof(*fdm) )
			{
				PutPos(15, "Start Cyl. Write Precompensation", COLUMN2, Fmt(t1,"%10u", SC16(fdm->start_cylwrt_precom)), 0);
				PutPos(15, "Start Cyl. Reduced Write Current", COLUMN2, Fmt(t1,"%10u", SC16(fdm->start_cylrdu_wrtcrt)), 0);
				PutPos(15, "Device Step Rate"                , COLUMN2, Fmt(t1,"%10u", SC16(fdm->dev_step_rate      )), 0);
				PutPos(15, "Device Step Plus Width"          , COLUMN2, Fmt(t1,"%10u", fdm->dev_step_pluswidth       ), 0);
				PutPos(15, "Head Settle Delay"               , COLUMN2, Fmt(t1,"%10u", SC16(fdm->head_settle_delay  )), 0);
				PutPos(15, "Motor On Delay"                  , COLUMN2, Fmt(t1,"%10u", fdm->motor_on_delay           ), 0);
				PutPos(15, "Motor Off Delay"                 , COLUMN2, Fmt(t1,"%10u", fdm->motor_off_delay          ), 0);
				PutPos(15, "Motor ON"                        , COLUMN2, Fmt(t1,"%10s", GetOnOff(fdm->motor_on       )), 0);
				PutPos(15, "Start Sector Number"             , COLUMN2, Fmt(t1,"%10u", fdm->start_sec_num            ), 0);
				PutPos(15, "True Ready"                      , COLUMN2, Fmt(t1,"%10s", GetOnOff(fdm->true_ready     )), 0);
				PutPos(15, "Step Plus per Cylinder"          , COLUMN2, Fmt(t1,"%10u", fdm->step_plus_per_cyl        ), 0);
				PutPos(15, "Write Compensation"              , COLUMN2, Fmt(t1,"%10u", fdm->write_compensation       ), 0);
				PutPos(15, "Head Load Delay"                 , COLUMN2, Fmt(t1,"%10u", fdm->head_load_delay          ), 0);
				PutPos(15, "Head Unload Delay"               , COLUMN2, Fmt(t1,"%10u", fdm->head_unload_delay        ), 0);
				switch ( fdm->pin34 )
				{
					case  0: s = "Open"        ; break;
					case  1: s = "Ready"       ; break;
					case  2: s = "Disk changed"; break;
					default: s = ""            ; break;
				}
				PutPos(15, "PIN34"                           , COLUMN2+9, Fmt(t1,"%1XH %s", fdm->pin34, s), 0);
				switch ( fdm->pin4 )
				{
					case  0: s = "Open"     ; break;
					case  1: s = "In Use"   ; break;
					case  2: s = "Eject"    ; break;
					case  4: s = "Head load"; break;
					default: s = ""         ; break;
				}
				PutPos(15, "PIN4"                            , COLUMN2+9, Fmt(t1,"%1XH %s", fdm->pin4, s), 0);
				switch ( fdm->pin1 )
				{
					case  0: s = "Open"             ; break;
					case  1: s = "Disk Change reset"; break;
					default: s = ""                 ; break;
				}
				PutPos(15, "PIN1"                            , COLUMN2+9, Fmt(t1,"%1XH %s", fdm->pin1, s), 0);
				PutPos(15, "PIN2"                            , COLUMN2+9, Fmt(t1,"%1XH", fdm->pin2), 0);
				PutPos(15, "Medium Rotation Rate"            , COLUMN2, Fmt(t1,"%10u",SC16(fdm->medium_rot_rate)), 0);
			}
				break;
			}
			case 0x08:	// Cache Control Parameter
			{
				if ( p[1] < 10 )
					goto ERR;
				fprintf(stdout, "    %-33s - Cache Control Parameter\n", msg);

				PutPos(15, TEXT("Write Cache Enable"               ), COLUMN2+5, Fmt(t1,"%5s", GetOnOff((p[2]>>2) & 1)), 0);
				PutPos(15, TEXT("Multiplication Factor"            ), COLUMN2+5, Fmt(t1,"%5s", GetOnOff((p[2]>>1) & 1)), 0);
				PutPos(15, TEXT("Read Cache Disable"               ), COLUMN2+5, Fmt(t1,"%5s", GetOnOff(p[2] & 1)     ), 0);
				PutPos(15, TEXT("Demand Read Retention Priority"   ), COLUMN2  , Fmt(t1,"%10u",(p[3]>>4) & 0xF        ), 0);
				PutPos(15, TEXT("Write Retention Priority"         ), COLUMN2  , Fmt(t1,"%10u", p[3] & 0xF            ), 0);
				PutPos(15, TEXT("Disable Pre-fetch Transfer Length"), COLUMN2  , Fmt(t1,"%10u", SC16(p+4)             ), 0);
				PutPos(15, TEXT("Minimum Pre-fetch"                ), COLUMN2  , Fmt(t1,"%10u", SC16(p+6)             ), 0);
				PutPos(15, TEXT("Maximum Pre-fetch"                ), COLUMN2  , Fmt(t1,"%10u", SC16(p+8)             ), 0);
				PutPos(15, TEXT("Maximum Pre-fetch Ceiling"        ), COLUMN2  , Fmt(t1,"%10u", SC16(p+10)            ), 0);
				break;
			}
			case 0x09:	// Periphery Device Parameter
			{
				if ( p[1] < 7 )
					goto ERR;
				fprintf(stdout, "    %-33s - Periphery Device Parameter\n", msg);
				sprintf(s1,"%04X", (p[2]<<8) + p[3]);
				fprintf(stdout, "               %-33s %10s\n", "interface ID", s1);
				this->PrintDumpOct("          vender depended", TRUE, p[1]-6, &p[8]);
				break;
			}
			case 0x0A: // Control Mode Parameter
			{
				DevCtl::ScsiPagecode0A_Cmp	*cmp;

				if ( p[1] + 2 < 6 )
					goto ERR;
				fprintf(stdout, "    %-33s - Control Mode Parameter\n", msg);

				cmp = (DevCtl::ScsiPagecode0A_Cmp *) p;
				PutPos(15, "Report Log Exception Condition"    , COLUMN2, Fmt(t1, "%10s", GetOnOff(cmp->rlec   )), 0);
				PutPos(15, "Global Logging Target Save Disable", COLUMN2, Fmt(t1, "%10s", GetOnOff(cmp->gltsd  )), 0);
				PutPos(15, "Descriptor Format Sense Data"      , COLUMN2, Fmt(t1, "%10s", GetOnOff(cmp->d_sense)), 0);
				PutPos(15, "Task Set Type"                     , COLUMN2, Fmt(t1, "%10u", cmp->task_set_type    ), 0);
				PutPos(15, "Queue Error Managemnt"             , COLUMN2, Fmt(t1, "%10u", cmp->queue_err_man    ), 0);
				PutPos(15, "Queue Algorithm Modifier"          , COLUMN2, Fmt(t1, "%10u", cmp->queue_alg_mod    ), 0);
				PutPos(15, "Software Write Protected"          , COLUMN2, Fmt(t1, "%10s", GetOnOff(cmp->write_protect)), 0);
				PutPos(15, "Unit Attention Interlocks control" , COLUMN2, Fmt(t1, "%10u", cmp->ua_intlck_ctrl   ), 0);
				PutPos(15, "Report Check"                      , COLUMN2, Fmt(t1, "%10s", GetOnOff(cmp->report_check)), 0);
				PutPos(15, "Task Aborted Status"               , COLUMN2, Fmt(t1, "%10s", GetOnOff(cmp->task_abort_stt)), 0);
				PutPos(15, "Autoload Mode"                     , COLUMN2, Fmt(t1, "%10u", cmp->autoload_mode     ), 0);
				PutPos(15, "Application Tag Owner"             , COLUMN2, Fmt(t1, "%10s", GetOnOff(cmp->app_tag_owner)), 0);
				if ( cmp->length + 2 >= 12 )
				{
					PutPos(15, "Busy Timeout Period"               , COLUMN2+6, Fmt(t1, "%04XH", SC16(cmp->busy_timeout  )), 0);
					PutPos(15, "Extended Self-Test Completion Time", COLUMN2+6, Fmt(t1, "%04XH", SC16(cmp->ext_stest_time)), 0);
				}
				break;
			}
			case 0x0B:	// Media Type Parameter
			{
				if ( p[1] < 6 )
					goto ERR;
				PutPos(4,msg, 38," - Media Type Parameter", 0);	// XXH XXH XXH XXH
				PutPos(15, "Media Type 1-4", COLUMN2-5, Fmt(t1,"%02XH %02XH %02XH %02XH", p[4],p[5],p[6],p[7]), 0);
				break;
			}
			case 0x0C:	// Notch/Partition Parameter
			{
				if ( p[1] + 2 < 24 )
					goto ERR;
				PutPos(4,msg, 38," - Notch/Partition Parameter", 0);
				PutPos(15, "Nocthed Device"            , COLUMN2, Fmt(t1,"%10s", GetOnOff(p[2] & 0x80)), 0);
				PutPos(15, "Logical or Physical Notch" , COLUMN2, Fmt(t1,"%10s", (p[2] & 0x40) ? "LOGICAL" : "PHYSICAL"), 0);
				PutPos(15, "Maximum Number of Notches" , COLUMN2, Fmt(t1,"%10u", (p[4]<<8)+p[5]), 0);
				PutPos(15, "Active Notch "             , COLUMN2, Fmt(t1,"%10u", (p[6]<<8)+p[7]), 0);
				PutPos(15, "Starting Boundary"         , COLUMN2, Fmt(t1,"%10u", (p[8]<<24)+(p[9]<<16)+(p[10]<<8)+p[11]), 0);
				PutPos(15, "Ending boundary"           , COLUMN2, Fmt(t1,"%10u", (p[12]<<24)+(p[13]<<16)+(p[14]<<8)+p[15]), 0);
				PutPos(15, "Pages Notched Parameter"   , COLUMN2-6, Fmt(t1,"%02X%02X%02X%02X%02X%02X%02X%02X",p[16],p[17],p[18],p[19],p[20],p[21],p[22],p[23]), 0);
				break;
			}
			case 0x0D:	// CD Device Parameter
			{
				char	*s;

				if ( p[1] < 6 )
					goto ERR;
				PutPos(4,msg, 38," - CD Device Parameters", 0);
				switch ( p[3] & 15 )
				{
					case  0: s = "Vender-specific"; break;
					case  1: s = "125ms"; break;
					case  2: s = "250ms"; break;
					case  3: s = "500ms"; break;
					case  4: s = "1s"; break;
					case  5: s = "2s"; break;
					case  6: s = "4s"; break;
					case  7: s = "8s"; break;
					case  8: s = "16s"; break;
					case  9: s = "32s"; break;
					case 10: s = "1min"; break;
					case 11: s = "2min"; break;
					case 12: s = "4min"; break;
					case 13: s = "8min"; break;
					case 14: s = "16min"; break;
					case 15: s = "32min"; break;
				}
				PutPos(15, "Inactivity Timer Multiplier", COLUMN2-5, Fmt(t1,"%15s", s), 0);
				PutPos(15, "MSF# - S Units Per MSF - M Unit", COLUMN2  , Fmt(t1,"%10u", (p[4]<<8) + p[5]), 0);
				PutPos(15, "MSF# - F Units Per MSF - S Unit", COLUMN2  , Fmt(t1,"%10u", (p[6]<<8) + p[7]), 0);
				break;
			}
			case 0x1A:	// Power Condition Page
			{
				if ( p[1] < 10 )
					goto ERR;
				PutPos(4,msg, 38," - Power Condition", 0);
				PutPos(15, "Idle"                  , COLUMN2+7, Fmt(t1,"%3s", GetOnOff(p[2] & 0x02)), 0);
				PutPos(15, "Standby"               , COLUMN2+7, Fmt(t1,"%3s", GetOnOff(p[2] & 0x01)), 0);
				PutPos(15, "Idle Timer"            , COLUMN2  , Fmt(t1,"%10u", (p[ 4]<<24)+(p[ 5]<<16)+(p[ 6]<<8)+p[ 7]), 0);
				PutPos(15, "Standby Timer"         , COLUMN2  , Fmt(t1,"%10u", (p[ 8]<<24)+(p[ 9]<<16)+(p[10]<<8)+p[11]), 0);
				break;
			}
			case 0x2A:	// MM Capabilities & Mechanical Status
				PutPos(4,msg, 38," - MM Capabilities & Mechanical Status", 0);
				if ( this->PrintPagecode2A(devParam, p[1] + 2, p ) < 0 )
					goto ERR;
				break;

			default:
			{
			ERR:
				this->PrintDumpOct(msg, TRUE, p[1], p+2);
				break;
			}
		}
		p += (unsigned)p[1] + 2;
	}

	return 0;
}

// 2A: MM Capabilities & Mechanical Status
int		DevTest::PrintPagecode2A(DevTest::DevParam & devParam, unsigned size, void * buffer)
{
	DevCtl::ScsiPagecode2A_Mms	*mms;
	const char					*s;
	char						t1[256], t2[256];
	unsigned					len, base;

	mms = (DevCtl::ScsiPagecode2A_Mms *) buffer;
	if ( size < 22 )
		return -1;
	base = 70;

	PutPos(15, "Read"
	      ,22, "CD-R"   , COLUMN2, Fmt(t1,"%10s", GetOnOff(mms->rd.cdr   )), 0);
	PutPos(22, "CD-RW"  , COLUMN2, Fmt(t1,"%10s", GetOnOff(mms->rd.cdrw  )), 0);
	PutPos(22, "DVD-ROM", COLUMN2, Fmt(t1,"%10s", GetOnOff(mms->rd.dvdrom)), 0);
	PutPos(22, "DVD-R"  , COLUMN2, Fmt(t1,"%10s", GetOnOff(mms->rd.dvdr  )), 0);
	PutPos(22, "DVD-RAM", COLUMN2, Fmt(t1,"%10s", GetOnOff(mms->rd.dvdram)), 0);

	PutPos(15, "Write"
	      ,22, "CD-R"   , COLUMN2, Fmt(t1,"%10s", GetOnOff(mms->wr.cdr   )), 0);
	PutPos(22, "CD-R/RW", COLUMN2, Fmt(t1,"%10s", GetOnOff(mms->wr.cdrrw )), 0);
	PutPos(22, "Test"   , COLUMN2, Fmt(t1,"%10s", GetOnOff(mms->wr.test  )), 0);
	PutPos(22, "DVD-R"  , COLUMN2, Fmt(t1,"%10s", GetOnOff(mms->wr.dvdr  )), 0);
	PutPos(22, "DVD-RAM", COLUMN2, Fmt(t1,"%10s", GetOnOff(mms->wr.dvdram)), 0);

	PutPos(15, "Audio Play"                       , COLUMN2, Fmt(t1,"%10s", GetOnOff(mms->audio_play   )), 0);
	PutPos(15, "Composite Audio/Video Data Stream", COLUMN2, Fmt(t1,"%10s", GetOnOff(mms->composite    )), 0);
	PutPos(15, "Digital Port (1)"                 , COLUMN2, Fmt(t1,"%10s", GetOnOff(mms->digital_port1)), 0);
	PutPos(15, "Digital Port (2)"                 , COLUMN2, Fmt(t1,"%10s", GetOnOff(mms->digital_port2)), 0);
	PutPos(15, "Mode 2 From 1"                    , COLUMN2, Fmt(t1,"%10s", GetOnOff(mms->mode2_from1  )), 0);
	PutPos(15, "Mode 2 From 2"                    , COLUMN2, Fmt(t1,"%10s", GetOnOff(mms->mode2_from2  )), 0);
	PutPos(15, "Multi Session"                    , COLUMN2, Fmt(t1,"%10s", GetOnOff(mms->multi_session)), 0);
	PutPos(15, "Buffer Under-run Free Recording"  , COLUMN2, Fmt(t1,"%10s", GetOnOff(mms->buf          )), 0);

	PutPos(15, "CD-DA Command supported"              , COLUMN2+7, Fmt(t1,"%3s", GetOnOff(mms->cdda_cmd        )), 0);
	PutPos(15, "CD-DA Stream is Accurate"             , COLUMN2+7, Fmt(t1,"%3s", GetOnOff(mms->cdda_stream_acc )), 0);
	PutPos(15, "Sub-channels R-W Reading Supported"   , COLUMN2+7, Fmt(t1,"%3s", GetOnOff(mms->rw_supported    )), 0);
	PutPos(15, "R-W Sub-ch De-interleaved & Corrected", COLUMN2+7, Fmt(t1,"%3s", GetOnOff(mms->rw_deinterleaved)), 0);
	PutPos(15, "C2 Error Pointers and C2 Block Error" , COLUMN2+7, Fmt(t1,"%3s", GetOnOff(mms->c2_pointer      )), 0);
	PutPos(15, "International Standard Recording Code", COLUMN2+7, Fmt(t1,"%3s", GetOnOff(mms->isrc            )), 0);
	PutPos(15, "Media Catalog #(Uniform Product Code)", COLUMN2+7, Fmt(t1,"%3s", GetOnOff(mms->upc             )), 0);
	PutPos(15, "Disc Bar Code"                        , COLUMN2+7, Fmt(t1,"%3s", GetOnOff(mms->read_bar_code   )), 0);

	PutPos(15, "Lock"                , COLUMN2+7, Fmt(t1,"%3s" , GetOnOff(mms->lock          )), 0);
	PutPos(15, "Lock Status"         , COLUMN2  , Fmt(t1,"%10s", mms->lock_status ? "LOCKED" : "UNLOCKED"), 0);
	PutPos(15, "Prevent/Allow Jumper", COLUMN2+7, Fmt(t1,"%3s" , GetOnOff(mms->prevend_jumper)), 0);
	PutPos(15, "Eject"               , COLUMN2+7, Fmt(t1,"%3s" , GetOnOff(mms->eject         )), 0);

	static const char	*loading_mecha_tbl[8] =
	{
		"Caddy type loading mechanism",
		"Tray type loading mechanism",
		"Popup type loading mechanism",
		"Loading Mechanism Type 3",
		"Changer with individually changeable discs",
		"Changer using a Magazine mechanism",
		"Type 6",
		"Type 7",
	};
	s = loading_mecha_tbl[mms->loading_type];
	len = strlen(s);
	if ( COLUMN2+10-len > 15 + 24 )
	{
		PutPos(15, "Loading Mechanism Type", COLUMN2+10-len, s, 0);
	} else
	{
		PutPos(15, "Loading Mechanism Type", 0);
		PutPos(COLUMN2+10-len, s, 0);
	}
	PutPos(15, "Separate Audio Volume Levels"  , COLUMN2+7, Fmt(t1,"%3s" , GetOnOff(mms->sep_vol_level   )), 0);
	PutPos(15, "Separate Audio Channel Mute"   , COLUMN2+7, Fmt(t1,"%3s" , GetOnOff(mms->sep_ch_mute     )), 0);
	PutPos(15, "Changer Supports Disc Present" , COLUMN2+7, Fmt(t1,"%3s" , GetOnOff(mms->changer_support )), 0);
	PutPos(15, "Software Slot Selection"       , COLUMN2+7, Fmt(t1,"%3s" , GetOnOff(mms->sss             )), 0);
	PutPos(15, "Side Change Capable"           , COLUMN2+7, Fmt(t1,"%3s" , GetOnOff(mms->side_change_capa)), 0);
	PutPos(15, "R through W in Lead-in"        , COLUMN2+7, Fmt(t1,"%3s" , GetOnOff(mms->rw_in_lead_in   )), 0);

	PutPos(15, "Maximum Read Speed"                , COLUMN2, Fmt(t1,"%10ukBps", SC16(mms->max_read )), 0);
	PutPos(15, "Number of Volume Levels Supportted", COLUMN2, Fmt(t1,"%10u", SC16(mms->vol_level)), 0);
	PutPos(15, "Buffer size Supported"             , COLUMN2, Fmt(t1,"%10uKB", SC16(mms->buf_size )), 0);
	PutPos(15, "Current Read Speed"                , COLUMN2, Fmt(t1,"%10ukBps", SC16(mms->cur_read )), 0);

	static const char *length_tbl[4] =
	{
		"32",
		"16",
		"24",
		"24(I2c)",
	};
	PutPos(15, "BCKF"       , COLUMN2+7, Fmt(t1,"%3s" , GetOnOff(mms->bckf)), 0);
	PutPos(15, "RCK"        , COLUMN2+7, Fmt(t1,"%3s" , GetOnOff(mms->rck )), 0);
	PutPos(15, "LSBF"       , COLUMN2+7, Fmt(t1,"%3s" , GetOnOff(mms->lsbf)), 0);
	PutPos(15, "Word Length", COLUMN2  , Fmt(t1,"%10s", length_tbl[mms->desc_bit_length]), 0);

	PutPos(15, "Maximum Write Speed"           , COLUMN2, Fmt(t1,"%10ukBps", SC16(mms->max_write)), 0);
	PutPos(15, "Current Write Speed"           , COLUMN2, Fmt(t1,"%10ukBps", SC16(mms->cur_write)), 0);

	if ( size < 32 )
		return 0;

	unsigned	n, num_tbl = SC16(mms->num_desc_tbl);
	BYTE		*d;

	PutPos(15, "Copy Management Revision Supported"    , COLUMN2+6, Fmt(t1,"%04XH", SC16(mms->copy_man_rev )), 0);
	static const char *	rot_ctrl_tbl[] =
	{
		"CLV & Non-pure CAV",	// 00
		"Pure CAV"          ,	// 01
		"",
		"",
		"",
		"",
		"",
		"",
	};
	PutPos(15, "Rotation Control Selected"          , COLUMN2-8, Fmt(t1,"%18s", rot_ctrl_tbl[mms->rotation_ctrl]), 0);
	PutPos(15, "Current Write Speed Selected"       , COLUMN2  , Fmt(t1,"%10u" , SC16(mms->cur_write_sel        )), 0);
	PutPos(15, "Write Speed Performance Desciptor #", COLUMN2+4, Fmt(t1,"%6u"  , SC16(mms->num_desc_tbl         )), 0);
	d = (BYTE *)(mms+1);
	for ( n = 0; n < num_tbl; n++ )
	{
		PutPos(17, Fmt(t1,"Block #%3u %18s", n+1, rot_ctrl_tbl[d[1]&3])
		 , COLUMN2+5, Fmt(t2,"%5u", SC16(d+2)), 0);
		d += 4;
	}
	return 0;
}

int		DevTest::PrintDevInfo_defectData(DevTest::DevParam & devParam)
{
	unsigned	pathId, targetId, lun;
	BYTE		buffer[4096];
	int			list;

	if ( devParam.CheckInqData() == FALSE )
		return -1;
	pathId   = devParam.inqData->GetPathId();
	targetId = devParam.inqData->GetTargetId();
	lun      = devParam.inqData->GetLun();
	for ( list = 0; list < 2; list++ )
	{
		int			len, n;
		char		*textList, t1[80];
		BYTE		*p;

		if ( this->devCtl.ScsiReadDefectData(pathId, targetId, lun, 1<<list, 0, sizeof(buffer), buffer) < 0 )
			return -1;
		len = (buffer[2]<<8) + buffer[3];
#if	defined(DEVTEST_DUMP_CSTYLE)
		if ( this->opt.dumpCStyle )
		{
			fprintf(stdout, "defectData list:%u\n", list);
			this->PrintDumpCStyle(len + 2, buffer);
		}
#endif
		switch ( (buffer[1] >> 3) & 3 )
		{
			case  1: textList = "G List"    ; break;
			case  2: textList = "P List"    ; break;
			case  3: textList = "G & P List"; break;
			default: textList = ""          ; break;
		}
		//  ----+----1----+----2----+----3----+----4----+----5----+----6----+----7----+----8
		// "    Defect List (Block adress)                  G & P List "
		// "    <------------------------------- ########## ########## "
		p = &buffer[4];
		n = 0;
		switch ( buffer[1] & 7 )	// type
		{
			case 0:	// block address format
				sprintf(t1, "Defect List (Block Address)");
				fprintf(stdout, "    %-43s %s\n", t1, textList);
				while ( len > 0 )
				{
					fprintf(stdout,"    ");
					fprintf(stdout,"%02X%02X%02X%02X",
					    p[0], p[1], p[2], p[3]);
					if ( (n % 6) == 5 )
						fprintf(stdout,"\n");
					if ( len < 4 )
						break;
					++n;
					len -= 4;
					p += 4;
				}
				if ( n == 0 )
					fprintf(stdout,"        %s\n", "nothing");
				else if ( n % 6 )
					fprintf(stdout,"\n");
				break;
			case 4:	// index byte format
				sprintf(t1, "Defect List (Distance from index)");
				fprintf(stdout, "    %-43s %s\n", t1, textList);
				while ( len > 0 )
				{
					fprintf(stdout,"    ");
					fprintf(stdout,"Cy:%02X%02X%02X Hd:%02X Bt:%02X%02X%02X%02X",
					    p[0], p[1], p[2], p[3], p[4], p[5], p[6], p[7]);
					if ( (n % 3) == 2 )
						fprintf(stdout,"\n");
					if ( len < 8 )
						break;
					++n;
					len -= 8;
					p += 8;
				}
				if ( n == 0 )
					fprintf(stdout,"        %s\n", "nothing");
				else if ( n % 3 )
					fprintf(stdout,"\n");
				break;
			case 5:	// physical sector format
				sprintf(t1, "Defect List (Physical Sector Address)");
				fprintf(stdout, "    %-43s %s\n", t1, textList);
				while ( len > 0 )
				{
					fprintf(stdout,"    ");
					fprintf(stdout,"Cy:%02X%02X%02X Hd:%02X No:%02X%02X%02X%02X",
					    p[0], p[1], p[2], p[3], p[4], p[5], p[6], p[7]);
					if ( (n % 3) == 2 )
						fprintf(stdout,"\n");
					if ( len < 8 )
						break;
					++n;
					len -= 8;
					p += 8;
				}
				if ( n == 0 )
					fprintf(stdout,"        %s\n", "nothing");
				else if ( n % 3 )
					fprintf(stdout,"\n");
				break;
			case 6:	// vender unique
			default:
				sprintf(t1, "Defect List (Vender Unique)");
				fprintf(stdout, "    %-43s %s\n", t1, textList);
				if ( len == 0 )
					fprintf(stdout,"        %s\n", "nothing");
				else
					this->PrintDumpOct("", TRUE, len, p);
				break;
		}
	}
	return 0;
}

int		DevTest::PrintDevInfo_getConfiguration(DevTest::DevParam & devParam)
{
	unsigned	pathId, targetId, lun;
#if	defined(_MSC_VER)
#	pragma pack(push,1)
#endif
	struct	FeatureHead
	{
		BYTE	length[4];
		BYTE	rsv[2];
		BYTE	profile[2];
	} ATTB_PACKED;
	struct FeatureDesc
	{
		BYTE	featureCode[2];
		BYTE	pc;
		BYTE	add_len;
	} ATTB_PACKED;
#if	defined(_MSC_VER)
#	pragma pack(pop)
#endif
	FeatureHead	*fhead;
	FeatureDesc	*fdesc;
	BYTE		buffer[4096], *p;
	unsigned	length;
	char		t1[256], t2[256];

	if ( devParam.CheckInqData() == FALSE )
		return -1;
	pathId   = devParam.inqData->GetPathId();
	targetId = devParam.inqData->GetTargetId();
	lun      = devParam.inqData->GetLun();
	if ( this->devCtl.ScsiGetConfiguration(pathId, targetId, lun, 0x00, 0, sizeof(buffer), buffer) < 0 )
		return -1;
	fhead = (FeatureHead *) buffer;
	length = SC32(fhead->length);
	PutPos(4, Fmt(t1, "Configuration (length:%u) : Profile %04XH %s"
	  , length
	  , SC16(fhead->profile)
	  , DevTest::GetProfileDefinitions(t2, SC16(fhead->profile))
	  ), 0);
#if	0 // defined(DEVTEST_DUMP_CSTYLE)
	if ( this->opt.dumpCStyle )
	{
		fprintf(stdout, "GetConfiguration\n");
		this->PrintDumpCStyle(length + 4, buffer);
	}
#endif
	p = &buffer[8];
	while ( p < length + buffer + sizeof(fhead->length) )
	{
		unsigned	fcode, flen;
		char		*s;

		fdesc = (FeatureDesc *) p;
		p += sizeof(*fdesc);
		fcode = SC16(fdesc->featureCode);
		flen = fdesc->add_len;
		PutPos(6, Fmt(t1,"FeatureCode %04XH (length %3u) : %s"
		  , fcode
		  , fdesc->add_len
		  , DevTest::GetFeatureCodeName(t2, fcode)
		 ), 0);
		switch ( fcode )
		{
			case 0x0000:	// Profile List Feature
			{
#if	defined(_MSC_VER)
#	pragma pack(push,1)
#endif
				struct ProfileList
				{
					BYTE	profileNumber[2];
					BYTE	pc;
					BYTE	add_len;
				} ATTB_PACKED;
#if	defined(_MSC_VER)
#	pragma pack(pop)
#endif
				ProfileList	*plist;
				unsigned		pn;

				while ( p < ((BYTE *)fdesc) + sizeof(*fdesc) + fdesc->add_len )
				{
					plist = (ProfileList *) p;
					pn = SC16(plist->profileNumber); 
					PutPos(15, Fmt(t1,"Profile Number %04XH %s"
					  , pn
					  , DevTest::GetProfileDefinitions(t2, pn)
					),0);
					p += sizeof(*plist);
				}
				break;
			}
			case 0x0001:	// Core Feature
			{
				DWORD	pis;
				if ( fdesc->add_len < 4 )
					goto DUMP;
				pis = SC32(p);
				switch ( pis )
				{
					case 0x00000000: s = "Unspecified"   ; break;
					case 0x00000001: s = "SCSI Family"   ; break;
					case 0x00000002: s = "ATAPI"         ; break;
					case 0x00000003: s = "IEEE 1394-1995"; break;
					case 0x00000004: s = "IEEE 1394A"    ; break;
					case 0x00000005: s = "Fibre Channel" ; break;
					case 0x00000006: s = "IEEE 1394B"    ; break;
					case 0x00000007: s = "Serial ATA"    ; break;
					case 0x00000008: s = "USB"           ; break;
					default: s = ""; break;
				}
				PutPos(15, "Physical Interface Standard", COLUMN2+2, Fmt(t1,"%08XH %s", pis, s),0);
				break;
			}
			case 0x0003:	// Removable Medium Feature
			{
				BYTE		*data;
				const char	*msg;
				unsigned	len;

				static const char	*loading_mecha_tbl[8] =
				{
					"Caddy/Slot type loading mechanism",
					"Tray type loading mechanism",
					"Embedded changer with individually changeable discs",
					"type 3",
					"Embedded changer using a magazine mechanism",
					"type 5",
					"type 6",
					"type 7",
				};

				if ( fdesc->add_len < 4 )
					goto DUMP;
				data = p;
				PutPos(15, "Lock"          , COLUMN2+7, Fmt(t1,"%3u", data[0] & 1),0);
				PutPos(15, "Prevent Jumper", COLUMN2+7, Fmt(t1,"%3u", (data[0]>>2) & 1),0);
				PutPos(15, "Eject"         , COLUMN2+7, Fmt(t1,"%3u", (data[0]>>3) & 1),0);
				msg = loading_mecha_tbl[data[0]>>5];
				len = strlen(msg);
				if ( COLUMN2+10-len > 15 + 24 )
				{
					PutPos(15, "Loading Mechanism Type", COLUMN2+10-len, msg, 0);
				} else
				{
					PutPos(15, "Loading Mechanism Type", 0);
					PutPos(COLUMN2+10-len, msg, 0);
				}
				goto DUMP;
				break;
			}
			default:
			DUMP:
				this->PrintDumpOct("", TRUE, fdesc->add_len, p);
				break;
		} // end of switch ( fcode )
		p = ((BYTE *)fdesc) + sizeof(*fdesc) + fdesc->add_len;
	} // end of while (p < )

	return 0;
}

char *	DevTest::GetFeatureCodeName(char * buffer, unsigned fcode) // static
{
	const char	*s;

	switch ( fcode )
	{
		case 0x0000: s = "Profile List Feature"                  ; break;
		case 0x0001: s = "Core Feature"                          ; break;
		case 0x0002: s = "Morphing Feature"                      ; break;
		case 0x0003: s = "Removable Medium Feature"              ; break;
		case 0x0004: s = "Write Protect Feature"                 ; break;
		case 0x0010: s = "Random Readable Feature"               ; break;
		case 0x001D: s = "Multi-Read Feature"                    ; break;
		case 0x001E: s = "CD Read Feature"                       ; break;
		case 0x001F: s = "DVD Read Feature"                      ; break;
		case 0x0020: s = "Random Writable Feature"               ; break;
		case 0x0021: s = "Incremental Streaming Writable"        ; break;
		case 0x0022: s = "Sector Erasable Feature"               ; break;
		case 0x0023: s = "Formattable Feature"                   ; break;
		case 0x0024: s = "Hardware Defect Management Feature"    ; break;
		case 0x0025: s = "Write Once Feature"                    ; break;
		case 0x0026: s = "Restricted Overwrite Feature"          ; break;
		case 0x0027: s = "CD-RW CAV Write Feature"               ; break;
		case 0x0028: s = "The MRW Feature"                       ; break;
		case 0x0029: s = "Enhanced Defect Reporting Feature"     ; break;
		case 0x002A: s = "DVD+RW Feature"                        ; break;
		case 0x002B: s = "DVD+R Feature"                         ; break;
		case 0x002C: s = "Rigid Restricted Overwrite Feature"    ; break;
		case 0x002D: s = "CD Track at Once Feature"              ; break;
		case 0x002E: s = "CD Mastering (Session at Once) Feature"; break;
		case 0x002F: s = "DVD-R/-RW Write Feature"               ; break;
		case 0x0030: s = "Double Density CD Read Feature"        ; break;
		case 0x0031: s = "Double Density CD Write Feature"       ; break;
		case 0x0032: s = "Double Density CD-RW Write Feature"    ; break;
		case 0x0033: s = "Layer Jump Recording Feature"          ; break;	// Add 2005.12.19
		case 0x0037: s = "CD-RW Media Write Support Feature"     ; break;
		case 0x0038: s = "BD-R Pseudo-Overwrite Feature"         ; break;	// Add 2005.12.19
		case 0x003A: s = "DVD+RW Dual Layer Feature"             ; break;	// Add 2005.12.19
		case 0x003B: s = "DVD+R Double Layer Feature"            ; break;
		case 0x0040: s = "BD Read Feature"                       ; break;	// Add 2005.12.19
		case 0x0041: s = "BD Write Feature"                      ; break;	// Add 2005.12.19
		case 0x0042: s = "TSR (Timely Safe Recording) Feature"   ; break;	// Add 2005.12.19
		case 0x0050: s = "HD DVD Read Feature"                   ; break;	// Add 2005.12.19
		case 0x0051: s = "HD DVD Write Feature"                  ; break;	// Add 2005.12.19
		case 0x0100: s = "Power Management Feature"              ; break;
		case 0x0101: s = "S.M.A.T. Feature"                      ; break;
		case 0x0102: s = "Embedded Changer Feature"              ; break;
		case 0x0103: s = "CD Audio External Play Feature"        ; break;
		case 0x0104: s = "Microcode Upgrade Feature"             ; break;
		case 0x0105: s = "Timeout Feature"                       ; break;
		case 0x0106: s = "DVD CSS Feature"                       ; break;
		case 0x0107: s = "Real Time Steaming Feature"            ; break;
		case 0x0108: s = "Logical Unit Serial Number Feature"    ; break;
		case 0x0109: s = "Media Serial Number Feature"           ; break;
		case 0x010A: s = "Disc Control Blocks"                   ; break;
		case 0x010B: s = "DVD CPRM"                              ; break;
		case 0x010C: s = "Firmware Information"                  ; break;
		case 0x010D: s = "AACS Feature"                          ; break;	// Add 2005.12.19
		case 0x0110: s = "VCPS Feature"                          ; break;	// Add 2005.12.19
		default: s = ""; break;
	}
	strcpy( buffer, s);
	return buffer;
}

char *	DevTest::GetProfileDefinitions(char * buffer, unsigned profileNumber)
{
	const char	*s;

	switch ( profileNumber )
	{
		case 0x0000: s = "No Current Profile"         ; break;
		case 0x0001: s = "Non-Removable Profile"      ; break;
		case 0x0002: s = "Removable Disk"             ; break;
		case 0x0003: s = "Magneto-Optical Erasable"   ; break;
		case 0x0004: s = "Optical Write Once"         ; break;
		case 0x0005: s = "AS-MO"                      ; break;
		case 0x0008: s = "CD-ROM"                     ; break;
		case 0x0009: s = "CD-R"                       ; break;
		case 0x000A: s = "CD-RW"                      ; break;
		case 0x0010: s = "DVD-ROM"                    ; break;
		case 0x0011: s = "DVD-R Sequential Recording" ; break;
		case 0x0012: s = "DVD-RAM"                    ; break;
		case 0x0013: s = "DVD-RW Restricted Overwrite"; break;
		case 0x0014: s = "DVD-RW Sequential Recording"; break;
		case 0x001A: s = "DVD+RW"                     ; break;
		case 0x001B: s = "DVD+R"                      ; break;
		case 0x0020: s = "DDCD-ROM"                   ; break;
		case 0x0021: s = "DDCD-R"                     ; break;
		case 0x0022: s = "DDCD-RW"                    ; break;
		case 0x002A: s = "DVD+RW Dual Layer"          ; break;	// Add 2005.12.19
		case 0x002B: s = "DVD+R Double Layer"         ; break;
		case 0x0040: s = "BD-ROM"                     ; break;	// Add 2005.12.19
		case 0x0041: s = "BD-R Sequential Recording"  ; break;	// Add 2005.12.19
		case 0x0042: s = "BD-R Random Recording"      ; break;	// Add 2005.12.19
		case 0x0043: s = "BD-RE Random Recording"     ; break;	// Add 2005.12.19
		case 0x0050: s = "HD DVD-ROM"                 ; break;	// Add 2005.12.19
		case 0x0051: s = "HD DVD-R"                   ; break;	// Add 2005.12.19
		case 0x0052: s = "HD DVD-RW"                  ; break;	// Add 2005.12.19
		default:
			s = "";
			break;
	}
	strcpy( buffer, s);
	return buffer;
}

int		DevTest::PrintDevInfo_dvdStructure(DevTest::DevParam & devParam)
{
	unsigned	pathId, targetId, lun;
	INT64		blkNum = 0;
	unsigned	blkLen = 0;
	DevCtl::SCSICAPACITYDATA	capaBuf;
	BYTE		buffer[4096], *p;
	char		t1[256], t2[256];
	unsigned	dslist_len, dsn, format, layer, layernum, dslen;
#if	defined(_MSC_VER)
#	pragma pack(push,1)
#endif
	struct tag_dslist
	{
		BYTE		length[2];
		BYTE		rsv[2];
		struct tag_entry
		{
			BYTE	format;
			BYTE	sdsrds;
			BYTE	length[2];
		} entry[256];
	} ATTB_PACKED dslist;
#if	defined(_MSC_VER)
#	pragma pack(pop)
#endif
	BOOL		dispTopMsg = FALSE;

	if ( devParam.CheckInqData() == FALSE )
		return -1;
	pathId   = devParam.inqData->GetPathId();
	targetId = devParam.inqData->GetTargetId();
	lun      = devParam.inqData->GetLun();

	if ( devParam.CheckCapacityData() )
	{
		capaBuf = *(devParam.GetCapacityData());
	} else
	{
		memset(&capaBuf, 0, sizeof(capaBuf));
		this->devCtl.ScsiReadCapacity( pathId, targetId, lun, sizeof(capaBuf), &capaBuf);
	}
	blkNum =           this->devCtl.ScsiGetLBA(capaBuf.logicalBlockAddress);
	blkLen = (unsigned)this->devCtl.ScsiGetLBA(capaBuf.blockLength);

	layernum = 1;
	for ( layer = 0; layer < layernum; layer++ )
	{
		memset( &dslist, 0, sizeof(dslist));
		format = 0xFF;
		if ( this->devCtl.ScsiReadDvdStructure(pathId, targetId, lun, layer, format, 0, sizeof(dslist), &dslist) < 0 )
			return -1;
		dslist_len = SC16(dslist.length);
		if ( dslist_len < 2 )
			continue;
#if	0 // defined(DEVTEST_DUMP_CSTYLE)
		if ( this->opt.dumpCStyle )
		{
			fprintf(stdout, "DVD Structure layer:%u\n", layer);
			this->PrintDumpCStyle(dslist_len + 2, (BYTE *)&dslist);
		}
#endif
		dslist_len -= 2;
		dsn = 0;
		for ( ; dslist_len >= 4; dsn++, dslist_len -= 4 )
		{
			BOOL		layer_use;
			unsigned	addr_use, addr, size, len;
			char		*desc, addr_buf[40];
			enum addr_enum { ADDR_OFF, ADDR_LBA, ADDR_PACK, ADDR_RMA, ADDR_CDESC, ADDR_PSN };

			format = dslist.entry[dsn].format;
			layer_use = FALSE;
			addr_use = 0;
			addr = 0;
			switch ( format )
			{
				case 0x00:
					desc = "Physical Format Information";
					layer_use = TRUE;
					break;
				case 0x01:
					desc = "DVD Copyright Information";
					layer_use = TRUE;
					break;
				case 0x02:
					desc = "DISC Key";
					break;
				case 0x03:
					desc = "BCA Information";
					break;
				case 0x04:
					desc = "DVD Disc Manufacturing Information";
					layer_use = TRUE;
					break;
				case 0x05:
					desc = "Copyright Management Information";
					addr_use = ADDR_LBA;
					break;
				case 0x06:
					desc = "Media Identifier";
					break;
				case 0x07: desc = "Media Key Block";
					break;
				case 0x08:
					desc = "DVD-RAM Disc Definition Structure";
					addr_use = ADDR_LBA;
					break;
				case 0x09:
					desc = "DVD-RAM Medium Status";
					break;
				case 0x0A:
					desc = "DVD-RAM Spare Area Information";
					break;
				case 0x0B:
					desc = "DVD-RAM Recording Type Information";
					break;
				case 0x0C:
					desc = "RMD in the last Border-out";
					break;
				case 0x0D:
					desc = "Recording Management Area Data";
					addr_use = ADDR_RMA;
					break;
				case 0x0E: desc = "Pre-recording Information in Lead-in";
					break;
				case 0x0F: desc = "Unique Disc Indentifier";
					break;
				case 0x10:
					desc = "Format Information of Control Data Zone in the Lead-in";
					layer_use = TRUE;
					break;
				case 0x11:
					desc = "ADIP Information";
					break;
				case 0x30:
					desc = "Disc Control Blocks";
					addr_use = ADDR_CDESC;
					break;
				case 0xC0:
					desc = "Write Protection Status";
					break;
				default:
					desc = ""; break;
			}
			if ( format == 0xFF || (layer > 0 && layer_use == FALSE) )
				continue;

			memset( buffer, 0, sizeof(buffer));
			size = (dslist.entry[dsn].length[0]<<8) + dslist.entry[dsn].length[1] + 4;
			if ( size > sizeof(buffer) )
				size = sizeof(buffer);
			if ( this->devCtl.ScsiReadDvdStructure(pathId, targetId, lun, layer, format, addr, size, buffer) < 0 )
				continue;
			p = buffer;
			dslen = (p[0]<<8) + p[1];
			if ( dispTopMsg == FALSE )
			{
				fprintf(stdout, "    Read DVD Structure\n");
				dispTopMsg = TRUE;
			}
			if ( addr_use > ADDR_OFF )
				sprintf(addr_buf, "%04u", addr);
			else
				strcpy(addr_buf, "----");
			PutPos(6, Fmt(t1,"Format:%02XH Layer:%1u  Address:%s  size:%4u :"
			  , format
			  , layer
			  , addr_buf
			  , dslen - 2
			  )
			, -1);
			len = strlen(desc);
			if ( 6 + 28 + 4 + 7 + 6 +len < 79 )
			{
				fprintf(stdout,"%s\n", desc);
			} else
			{
				fprintf(stdout,"\n");
				PutPos(79-len, desc, 0);
			}
#if	0 // defined(DEVTEST_DUMP_CSTYLE)
			if ( this->opt.dumpCStyle )
			{
				fprintf(stdout, "DVD Structure layer:%u Format:%04XH\n", layer, format);
				this->PrintDumpCStyle(dslen + 2, buffer);
			}
#endif

			switch ( format )
			{
				case 0x00:	// Physical Format Information
				{
					DevCtl::ScsiDvdStrFormat0000_Pfi	*pfi;
					const char		*s;
					unsigned		i, bit;

					if ( dslen-2 < sizeof(*pfi) )
						goto DUMP;
					p = &buffer[4];
					pfi = (DevCtl::ScsiDvdStrFormat0000_Pfi *)p;
					switch ( pfi->book_type )
					{
						case 0x00: s = "DVD-ROM"; break;
						case 0x01: s = "DVD-RAM"; break;
						case 0x02: s = "DVD-R"  ; break;
						case 0x03: s = "DVD-RW" ; break;
						case 0x09: s = "DVD+RW" ; break;
						case 0x0A: s = "DVD+R"  ; break;
						default  : s = ""       ; break;
					}
					PutPos(15, "Book Type"   , COLUMN2+9, Fmt(t1, "%1XH %s", pfi->book_type, s), 0);
					PutPos(15, "Part Version", COLUMN2  , Fmt(t1, "%10u", pfi->part_version), 0);
					switch ( pfi->disc_size )
					{
						case 0x00: s = "120mm"; break;
						case 0x01: s = "80mm" ; break;
						default  : s = ""     ; break;
					}
					PutPos(15, "Disc Size"   , COLUMN2  , Fmt(t1, "%10u %s", pfi->disc_size, s), 0);
					switch ( pfi->max_rate )
					{
						case 0x00: s = "2.52 Mbps"    ; break;
						case 0x01: s = "5.04 Mbps"    ; break;
						case 0x02: s = "10.08 Mbps"   ; break;
						case 0x0F: s = "Not Specified"; break;
						default  : s = ""             ; break;
					}
					PutPos(15, "Maximum Rate", COLUMN2+9, Fmt(t1, "%1XH %s", pfi->max_rate, s), 0);
					PutPos(15, "Number of Layers", COLUMN2  , Fmt(t1, "%10u", pfi->num_layer), 0);
					PutPos(15, "Track Path"      , COLUMN2  , Fmt(t1, "%10u", pfi->track_path), 0);
					PutPos(15, "Layer Type"      , COLUMN2+9, Fmt(t1, "%1XH", pfi->layer_type), 0);
					for ( i = 0; i < 4; i++ )
					{
						bit = 1 << i;
						switch ( i )
						{
							case  0: s = "Layer contains embossed data"  ; break;
							case  1: s = "Layer contains recordable data"; break;
							case  2: s = "Layer contains rewritable data"; break;
							default: s = NULL; break;
						}
						if ( s == NULL )
							continue;
						if ( pfi->layer_type & bit )
							PutPos(COLUMN2 + 10 - strlen(s), s, 0);
					}
					switch ( pfi->linear_density )
					{
						case 0x00: s = "0.267um/bit"      ; break;
						case 0x01: s = "0.293um/bit"      ; break;
						case 0x02: s = "0.409-0.435um/bit"; break;
						case 0x04: s = "0.280-0.291um/bit"; break;
						case 0x08: s = "0.353um/bit"      ; break;
						default  : s = ""                 ; break;
					}
					PutPos(15, "Linear Density", COLUMN2+9, Fmt(t1, "%1XH %s", pfi->linear_density, s), 0);
					switch ( pfi->track_density )
					{
						case 0x00: s = "0.74 um/track" ; break;
						case 0x01: s = "0.80 um/track" ; break;
						case 0x02: s = "0.615um/track"; break;
						default  : s = ""              ; break;
					}
					PutPos(15, "Track Density"                      , COLUMN2+9, Fmt(t1, "%1XH %s", pfi->track_density, s), 0);
					PutPos(15, "Starting Phy. Sector # of Data Area", COLUMN2+4, Fmt(t1, "%06XH", SC24(pfi->start_sec_data)), 0);
					PutPos(15, "End Phy. Sector # of Data Area"     , COLUMN2+4, Fmt(t1, "%06XH", SC24(pfi->end_sec_data  )), 0);
					PutPos(15, "End Phy. Sector # in Layer 0"       , COLUMN2+4, Fmt(t1, "%06XH", SC24(pfi->end_sec_data  )), 0);
					PutPos(15, "BCA"      , COLUMN2  , Fmt(t1, "%10u", pfi->bca), 0);

					this->DumpParam( 0, sizeof(*pfi), dslen-2, p);
					if ( layernum < (unsigned)pfi->num_layer + 1 )
						layernum = pfi->num_layer + 1;
					break;
				}
				case 0x09:	// DVD-RAM Medium Status
				{
					const char	*s;

					if ( dslen-2 < 4 )
						goto DUMP;
					p = &buffer[4];
					PutPos(15, TEXT("Medium in Cartridge"                   ), COLUMN2+9, Fmt(t1, TEXT("%1u"), (p[0]>>7) & 1), 0);
					PutPos(15, TEXT("Medium Out from Cartridge"             ), COLUMN2+9, Fmt(t1, TEXT("%1u"), (p[0]>>6) & 1), 0);
					PutPos(15, TEXT("Medium Specific Write Inhibition"      ), COLUMN2+9, Fmt(t1, TEXT("%1u"), (p[0]>>3) & 1), 0);
					PutPos(15, TEXT("Medium Cartrige Write Protection"      ), COLUMN2+9, Fmt(t1, TEXT("%1u"), (p[0]>>2) & 1), 0);
					PutPos(15, TEXT("Persistent Write Protection"           ), COLUMN2+9, Fmt(t1, TEXT("%1u"), (p[0]>>1) & 1), 0);
					switch ( p[1] )
					{
						case 0x00: s = "Disc shall not be written without a cartridge"; break;
						case 0x10: s = "Disc may be written without a cartridge"      ; break;
						default  : s = ""; break;
					}
					PutPos(15, TEXT("Disc Type Identification"              ), COLUMN2+8, Fmt(t1, TEXT("%02XH"), p[1]), 0);
					if ( (len = strlen(s)) > 0 )
						PutPos(COLUMN2+10-len, s, 0);
					PutPos(15, TEXT("DVD-RAM Specific Write Inhibition"     ), COLUMN2+8, Fmt(t1, TEXT("%02XH"), p[3]), 0);
					break;
				}
				case 0x0A:	// DVD-RAM Spare Area Information
				{
					int		blks;

					if ( dslen-2 < 12 )
						goto DUMP;
					p = &buffer[4];
					blks = SC32(p+0);
					PutPos(15, "Unused Primary Spare blocks"        , COLUMN2+1, Fmt(t1,"%9u (%8s)", blks,DevTest::FormatSize(t2,(INT64)blks*blkLen)), 0);
					blks = SC32(p+4);
					PutPos(15, "Unused Supplemntary Spare blocks"   , COLUMN2+1, Fmt(t1,"%9u (%8s)", blks,DevTest::FormatSize(t2,(INT64)blks*blkLen)), 0);
					blks = SC32(p+8);
					PutPos(15, "Allocated Supplemntary Spare blocks", COLUMN2+1, Fmt(t1,"%9u (%8s)", blks,DevTest::FormatSize(t2,(INT64)blks*blkLen)), 0);
					break;
				}
				case 0x0E:	// Pre-recording Information in Lead-in
				{
					DevCtl::ScsiDvdStrFomrta000E_Pil	*pil;

					if ( dslen-2 < sizeof(*pil) )
						goto DUMP;
					p = &buffer[4];
					pil = (DevCtl::ScsiDvdStrFomrta000E_Pil *)p;
				//	PutPos(15, "Filed ID"             , COLUMN2+8, Fmt(t1, "%02XH", pil->filed_id0), 0);
					PutPos(15, "Disc Application code", COLUMN2+8, Fmt(t1, "%02XH", pil->disc_appl_code), 0);
					PutPos(15, "Disc Physical code"   , COLUMN2+8, Fmt(t1, "%02XH", pil->disc_phy_code), 0);
					PutPos(15, "Last address of Data Recordable Area", COLUMN2+4, Fmt(t1, "%06XH", SC24(pil->last_addr)), 0);
					PutPos(15, "Part Version"         , COLUMN2+9, Fmt(t1, "%1XH", pil->part_version), 0);
					PutPos(15, "Extension code"       , COLUMN2+9, Fmt(t1, "%1XH", pil->ext_code), 0);
				//	PutPos(15, "Filed ID"             , COLUMN2+8, Fmt(t1, "%02XH", pil->filed_id8), 0);
					PutPos(15, "OPC Suggested Code"   , COLUMN2+8, Fmt(t1, "%02XH", pil->opc_sug_code), 0);
					PutPos(15, "Wavelength/OPC Suggested Code", COLUMN2+8, Fmt(t1, "%02XH", pil->wavelen_opc_code), 0);
					PutPos(15, "White Strategy Code"  , COLUMN2+2, Fmt(t1, "%08XH", SC32(pil->white_str_code)), 0);
				//	PutPos(15, "Filed ID"             , COLUMN2+8, Fmt(t1, "%02XH", pil->filed_id16), 0);
					PutPos(15, "Manufacturer ID"      , COLUMN2+4, Fmt(t1, "%6.6s", (char *)pil->manufacturer_id), 0);
				//	PutPos(15, "Filed ID"             , COLUMN2+8, Fmt(t1, "%02XH", pil->filed_id24), 0);
					PutPos(15, "Manufacturer ID"      , COLUMN2+4, Fmt(t1, "%6.6s", (char *)pil->manufacturer_id2), 0);
				//	PutPos(15, "Filed ID"             , COLUMN2+8, Fmt(t1, "%02XH", pil->filed_id32), 0);
				//	PutPos(15, "Manufacturer ID"      , COLUMN2+4, Fmt(t1, "%6.6s", (char *)pil->manufacturer_id3), 0);

					this->DumpParam( 0, sizeof(*pil), dslen-2, p);
					break;
				}
				case 0x0F:	// Unique Disc Identifier
				{
					if ( dslen-2 < 18 )
						goto DUMP;
					p = &buffer[4];
					PutPos(15, "Random Number", COLUMN2+6, Fmt(t1,"%04XH", (p[2]<<8)+p[3]), 0);
					PutPos(15, "Year/Month/Day hh:mm:ss", COLUMN2-9
					  , Fmt(t1,"%4.4s/%2.2s/%2.2s %2.2s:%2.2s:%2.2s"
					    , &p[ 4]
					    , &p[ 8]
					    , &p[10]
					    , &p[12]
					    , &p[14]
					    , &p[16]
					  ), 0);
					break;
				}
				default:
				DUMP:
					if ( dslen-2 < 16 )
						this->PrintDumpOct("", TRUE, dslen-2, p + 4);
					else
						this->DumpParam( 0, 0, dslen-2, p + 4);
					break;
			}
		} // end of for(dslist)
	} // end of for(layer)

	return 0;
}

int		DevTest::DumpParam(unsigned addr, unsigned off, unsigned bufsiz, BYTE * buffer)
{
	unsigned	size, x, y;
	BOOL		guide = 0;

	size = DevTest::GetNonZeroSize(bufsiz, buffer);
	if ( size < 16 )
	{
		if ( bufsiz >= 16 )
			size = 16;
		else
			size = bufsiz;
	}
	if ( (int)bufsiz - (int)size < 16 )
		size = bufsiz;
	for ( y = (off/16)*16; y < size; y += 16 )
	{
		if ( guide == 0 )
		{
			fprintf(stdout,"          OFFS +0 +1 +2 +3 +4 +5 +6 +7 +8 +9 +A +B +C +D +E +F ----------------\n");
			guide = 1;
		}
		fprintf(stdout,"          %04X", (addr+y) & 0xFFFF);
		for ( x = 0; x < 16; x++ )
		{
			unsigned	c;

			if ( y + x >= off && y + x < size )
			{
				c = buffer[y + x];
				fprintf(stdout, " %02X", c);
			} else
				fprintf(stdout, "   ");
		}
		fprintf(stdout," ");
		for ( x = 0; x < 16; x++ )
		{
			unsigned	c;

			if ( y + x >= off && y + x < size )
			{
				c = buffer[y + x];
				c = isprint(c) ? c : '.';
			} else
				c= ' ';
			fprintf(stdout,"%c", c);
		}
		fprintf(stdout,"\n");
	}
	if ( size < bufsiz )
	{
		char	tmp[80];

		PutPos(10, Fmt(tmp,"Remain %u bytes all zero", bufsiz-size), 0);
	}
	return 0;
}
