/////////////////////////////////////////////////////////////////////////
//	"devtest_ext2.cpp"
//------------------------------------------------------------------------
//	$Id: devtest_ext2.cpp 136 2004-09-28 12:28:08Z taro $
//////////////////////////////////////////////////////////////////////////

#include	<stdlib.h>
#include	<stdio.h>
#include	<string.h>
#include	<time.h>
#include	"devtest.h"
#include	"defs.h"

#include	"ext2lib.h"

int		DevTest::PrintExt2Info(DevTest::DevParam & devParam, INT64 addr, INT64 addrMax)
{
	Ext2Lib						ext2lib;
	Ext2Lib::ext2_super_block	sb;
	char						*s, t1[256],t2[80];
	unsigned					blkLen;
	INT64						blkNum;
	time_t						tmbuf;
	struct tm					*tp;

	if ( devParam.CheckInqData() == FALSE || devParam.CheckCapacityData() == FALSE )
		return -1;
	if ( ext2lib.OpenExt2(this->devCtl, *devParam.GetInqData(), *devParam.GetCapacityData(), addr, addrMax) < 0 )
		return -1;
	if ( ext2lib.CheckSuperBlock(sb) < 0 )
		return -1;
	blkLen = ext2lib.GetBlkSize();
	blkNum = ext2lib.GetBlkNum();
	// "  IBM HARD-DISK FORMAT    (########) ########## sectors   ####################B
	DevTest::fprintf(stdout, "  LINUX EXT2 FORMAT       (%8u) %10"FI64"u sectors   %20,I64uB\n"
	  , blkLen
	  , (INT64)blkNum
	  , (INT64)blkNum * (INT64)blkLen
	);
	PutPos(4, "Inodes count"         , COLUMN1, Fmt(t1,"%10u",(unsigned)Ext2Lib::ue32(sb.s_inodes_count     )) ,0);
	PutPos(4, "Blocks count"         , COLUMN1, Fmt(t1,"%10u",(unsigned)Ext2Lib::ue32(sb.s_blocks_count     )) ,0);
	PutPos(4, "Reserved blocks count", COLUMN1, Fmt(t1,"%10u",(unsigned)Ext2Lib::ue32(sb.s_r_blocks_count   )) ,0);
	PutPos(4, "Free blocks count"    , COLUMN1, Fmt(t1,"%10u",(unsigned)Ext2Lib::ue32(sb.s_free_blocks_count)) ,0);
	PutPos(4, "Free inodes count"    , COLUMN1, Fmt(t1,"%10u",(unsigned)Ext2Lib::ue32(sb.s_free_inodes_count)) ,0);
	PutPos(4, "First data Block"     , COLUMN1, Fmt(t1,"%10u",(unsigned)Ext2Lib::ue32(sb.s_first_data_block )) ,0);
	PutPos(4, "Log block size"       , COLUMN1, Fmt(t1,"%10u",(unsigned)Ext2Lib::ue32(sb.s_log_block_size   )) ,0);
	PutPos(4, "Log fragment size"    , COLUMN1, Fmt(t1,"%10d",     (int)Ext2Lib::se32(sb.s_log_frag_size    )) ,0);
	PutPos(4, "Blocks per group"     , COLUMN1, Fmt(t1,"%10u",(unsigned)Ext2Lib::ue32(sb.s_blocks_per_group )) ,0);
	PutPos(4, "Fragments per group"  , COLUMN1, Fmt(t1,"%10u",(unsigned)Ext2Lib::ue32(sb.s_frags_per_group  )) ,0);
	PutPos(4, "Inodes per group"     , COLUMN1, Fmt(t1,"%10u",(unsigned)Ext2Lib::ue32(sb.s_inodes_per_group )) ,0);
	tmbuf = (time_t)Ext2Lib::ue32(sb.s_mtime);
	tp = localtime(&tmbuf);
	PutPos(4, "Mount time", COLUMN1-9, Fmt(t1,"%4u/%02u/%02u %02u:%02u:%02u",tp->tm_year+1900,tp->tm_mon+1,tp->tm_mday,tp->tm_hour,tp->tm_min,tp->tm_sec) ,0);
	tmbuf = (time_t)Ext2Lib::ue32(sb.s_wtime);
	tp = localtime(&tmbuf);
	PutPos(4, "Write time", COLUMN1-9, Fmt(t1,"%4u/%02u/%02u %02u:%02u:%02u",tp->tm_year+1900,tp->tm_mon+1,tp->tm_mday,tp->tm_hour,tp->tm_min,tp->tm_sec) ,0);
	PutPos(4, "Mount count"          , COLUMN1, Fmt(t1,"%10u" ,(unsigned)Ext2Lib::ue16(sb.s_mnt_count        )) ,0);
	PutPos(4, "Max mount count"      , COLUMN1, Fmt(t1,"%10d" ,     (int)Ext2Lib::se16(sb.s_max_mnt_count    )) ,0);
	PutPos(4, "Magic signature"    , COLUMN1+6, Fmt(t1,"%04XH",(unsigned)Ext2Lib::ue16(sb.s_magic            )) ,0);
	PutPos(4, "File system state"  , COLUMN1+6, Fmt(t1,"%04XH",(unsigned)Ext2Lib::ue16(sb.s_state            )) ,0);
	if ( Ext2Lib::ue16(sb.s_state) & 0x0001 ) PutPos(COLUMN1-10, Fmt(t1,"%20s", "Unmounted cleanly") ,0);
	if ( Ext2Lib::ue16(sb.s_state) & 0x0002 ) PutPos(COLUMN1-10, Fmt(t1,"%20s", "Errors detected"  ) ,0);
	PutPos(4, "Minor revision level" , COLUMN1+6, Fmt(t1,"%04XH",(unsigned)Ext2Lib::ue16(sb.s_minor_rev_level)) ,0);
	tmbuf = (time_t)Ext2Lib::ue32(sb.s_lastcheck);
	tp = localtime(&tmbuf);
	PutPos(4, "Last check", COLUMN1-9, Fmt(t1,"%4u/%02u/%02u %02u:%02u:%02u",tp->tm_year+1900,tp->tm_mon+1,tp->tm_mday,tp->tm_hour,tp->tm_min,tp->tm_sec) ,0);
	PutPos(4, "Check interval (Day HH:MM:DD)", COLUMN1-4, Fmt(t1,"%14s", this->FormatTime(t2,Ext2Lib::ue32(sb.s_checkinterval))) ,0);
	if ( (s = ext2lib.GetOsName(Ext2Lib::ue32(sb.s_creator_os), t2)) == NULL )
		s = "";
	PutPos(4, "Creator OS"           , COLUMN1+2, Fmt(t1,"%08XH %s",(unsigned)Ext2Lib::ue32(sb.s_creator_os), s), 0);
	PutPos(4, "Revision level"       , COLUMN1+2, Fmt(t1,"%08XH",(unsigned)Ext2Lib::ue32(sb.s_rev_level      )) ,0);
	PutPos(4, "Reserved blocks UID"  , COLUMN1  , Fmt(t1,"%10u" ,(unsigned)Ext2Lib::ue16(sb.s_def_resuid     )) ,0);
	PutPos(4, "Reserved blocks GID"  , COLUMN1  , Fmt(t1,"%10u" ,(unsigned)Ext2Lib::ue16(sb.s_def_resgid     )) ,0);
	// 180 00:00:00
	
	return 0;
}
